package net.knarcraft.paidsigns.listener;

import net.knarcraft.knarlib.formatting.TranslatableMessage;
import net.knarcraft.paidsigns.PaidSigns;
import net.knarcraft.paidsigns.container.PaidSign;
import net.knarcraft.paidsigns.container.PaidSignConditionMatch;
import net.knarcraft.paidsigns.formatting.PaidSignsTranslatableMessage;
import net.knarcraft.paidsigns.manager.EconomyManager;
import net.knarcraft.paidsigns.manager.TrackedSignManager;
import net.knarcraft.paidsigns.utility.SignHelper;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.block.Block;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.SignChangeEvent;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

/**
 * A listener for listening to registered paid signs
 */
public class SignListener implements Listener {

    @EventHandler(priority = EventPriority.LOW)
    public void onSignChange(SignChangeEvent event) {
        if (event.isCancelled() || event.getPlayer().hasPermission("paidsigns.paymentexempt")) {
            return;
        }

        // As signs can be edited now, this event might be triggered on an already registered sign, so unregister
        if (SignHelper.isSign(event.getBlock())) {
            try {
                TrackedSignManager.removeTrackedSign(event.getBlock().getLocation(), true, false,
                        event.getSide());
            } catch (IOException ignored) {
            }
        }

        String[] lines = event.getLines();
        Map<String, PaidSign> allPaidSigns = PaidSigns.getInstance().getSignManager().getAllPaidSigns();

        //Generate the sorted list of the number of paid sign conditions matched for every paid sign
        List<PaidSignConditionMatch> allConditionMatches = new ArrayList<>();
        for (PaidSign paidSign : allPaidSigns.values()) {
            calculateConditionMatches(paidSign, lines, allConditionMatches);
        }
        allConditionMatches.sort(Collections.reverseOrder());

        //Make the player pay for the paid sign
        PaidSign mostExpensive = getMatchingSign(allConditionMatches);
        if (mostExpensive != null) {
            performPaidSignCheck(mostExpensive, event);
        }
    }

    /**
     * Gets the most specific paid sign match
     *
     * <p>This method chooses the most specific match, if any. If two matches have the same number of matched
     * conditions, it will choose the most expensive one.</p>
     *
     * @param allConditionMatches <p>A sorted list of all paid sign condition matches</p>
     * @return <p>The best matching paid sign, or null if no paid sign matches</p>
     */
    private PaidSign getMatchingSign(List<PaidSignConditionMatch> allConditionMatches) {
        PaidSign mostExpensive = null;
        int bestMatch = 1;
        for (PaidSignConditionMatch paidSignConditionMatch : allConditionMatches) {
            int conditionMatches = paidSignConditionMatch.conditionMatches();
            /* Stop if there is no matching paid signs. Also stop if a paid sign is encountered with fewer conditions 
            matching */
            if (conditionMatches >= bestMatch) {
                PaidSign paidSign = paidSignConditionMatch.paidSign();
                //Choose the most expensive paid sign between those with the same number of matched conditions
                if (mostExpensive == null || paidSign.getCost() > mostExpensive.getCost()) {
                    mostExpensive = paidSign;
                    bestMatch = conditionMatches;
                }
            } else {
                break;
            }
        }
        return mostExpensive;
    }

    /**
     * Calculates the number of conditions matching for the given paid sign
     *
     * <p>This calculates the number of matches, allowing the most specific paid sign to be chosen</p>
     *
     * @param paidSign            <p>The paid sign to calculate matches for</p>
     * @param lines               <p>The lines of a sign</p>
     * @param allConditionMatches <p>The number of conditions matched</p>
     */
    private void calculateConditionMatches(PaidSign paidSign, String[] lines, List<PaidSignConditionMatch> allConditionMatches) {
        if (paidSign.matches(lines)) {
            if (paidSign.matchAnyCondition()) {
                //For any match, it can only be assumed one lines matches
                allConditionMatches.add(new PaidSignConditionMatch(paidSign, 1));
            } else {
                //For a normal match, the number of conditions is equal to the number of matches
                allConditionMatches.add(new PaidSignConditionMatch(paidSign, paidSign.getConditions().size()));
            }
        } else {
            allConditionMatches.add(new PaidSignConditionMatch(paidSign, 0));
        }
    }

    /**
     * Performs necessary checks and performs the paid sign transaction
     *
     * @param paidSign <p>The paid sign to test against the sign lines</p>
     * @param event    <p>The triggered sign change event to cancel if necessary</p>
     */
    private void performPaidSignCheck(PaidSign paidSign, SignChangeEvent event) {
        Player player = event.getPlayer();
        String permission = paidSign.getPermission();
        //If a match is found, but the player is missing the permission, assume no plugin sign was created
        if (permission != null && !permission.trim().isEmpty() && !player.hasPermission(permission)) {
            return;
        }

        performPaidSignTransaction(paidSign, player, event);
        Location signLocation = event.getBlock().getLocation();

        if (!event.isCancelled()) {
            //Immediately refund if a plugin destroyed the sign within 5 ticks of the creation
            Bukkit.getScheduler().scheduleSyncDelayedTask(PaidSigns.getInstance(), () -> {
                Block block = signLocation.getBlock();
                if (!SignHelper.isSign(block)) {
                    try {
                        TrackedSignManager.removeTrackedSign(block.getLocation(), true, true);
                    } catch (IOException e) {
                        PaidSigns.getInstance().getLogger().log(Level.WARNING, String.format("Unable to save changes " +
                                "about removed tracked sign at %s", block.getLocation()));
                    }
                }
            }, 5);
        }
    }

    /**
     * Performs the transaction to pay for the paid sign
     *
     * @param paidSign <p>The paid sign a match has been found for</p>
     * @param player   <p>The player that created the sign</p>
     * @param event    <p>The sign change event that caused the sign to be created</p>
     */
    private void performPaidSignTransaction(PaidSign paidSign, Player player, SignChangeEvent event) {
        double cost = paidSign.getCost();
        boolean canAfford = EconomyManager.canAfford(player, cost);
        if (!canAfford) {
            PaidSigns.getStringFormatter().displayErrorMessage(player, replaceCost(cost, PaidSignsTranslatableMessage.ERROR_CANNOT_AFFORD));
            event.setCancelled(true);
        } else {
            EconomyManager.withdraw(player, cost);
            PaidSigns.getStringFormatter().displaySuccessMessage(player, replaceCost(cost, PaidSignsTranslatableMessage.SUCCESS_PAID_FOR_SIGN));
            try {
                TrackedSignManager.addTrackedSign(event.getBlock().getLocation(), player.getUniqueId(), cost, event.getSide());
            } catch (IOException ignored) {
            }
        }
    }

    /**
     * Replaces cost and unit placeholders in the given message
     *
     * @param cost    <p>The cost to insert</p>
     * @param message <p>The original message to replace the cost placeholder for</p>
     * @return <p>The message with the cost instead of the cost placeholder</p>
     */
    private String replaceCost(double cost, TranslatableMessage message) {
        return PaidSigns.getStringFormatter().replacePlaceholder(message,
                "{cost}", EconomyManager.format(cost));
    }

}
